@@ -181,11 +181,9 @@ import scipy.spatial.distance as distance
 from scipy._lib.six import string_types
 from scipy._lib.six import xrange
 
-_cpy_non_euclid_methods = {'single': 0, 'complete': 1, 'average': 2,
-                           'weighted': 6}
-_cpy_euclid_methods = {'centroid': 3, 'median': 4, 'ward': 5}
-_cpy_linkage_methods = set(_cpy_non_euclid_methods.keys()).union(
-    set(_cpy_euclid_methods.keys()))
+_LINKAGE_METHODS = {'single': 0, 'complete': 1, 'average': 2, 'centroid': 3,
+                    'median': 4, 'ward': 5, 'weighted': 6}
+_EUCLIDEAN_METHODS = ('centroid', 'median', 'ward')
 
 __all__ = ['ClusterNode', 'average', 'centroid', 'complete', 'cophenet',
            'correspond', 'cut_tree', 'dendrogram', 'fcluster', 'fclusterdata',
@@ -196,8 +194,12 @@ __all__ = ['ClusterNode', 'average', 'centroid', 'complete', 'cophenet',
            'to_mlab_linkage', 'to_tree', 'ward', 'weighted', 'distance']
 
 
+class ClusterWarning(UserWarning):
+    pass
+
+
 def _warning(s):
-    warnings.warn('scipy.cluster: %s' % s, stacklevel=3)
+    warnings.warn('scipy.cluster: %s' % s, ClusterWarning, stacklevel=3)
 
 
 def _copy_array_if_base_present(a):
@@ -253,6 +255,7 @@ def single(y):
     See Also
     --------
     linkage: for advanced creation of hierarchical clusterings.
+    scipy.spatial.distance.pdist : pairwise distance metrics
 
     """
     return linkage(y, method='single', metric='euclidean')
@@ -272,12 +275,13 @@ def complete(y):
     -------
     Z : ndarray
         A linkage matrix containing the hierarchical clustering. See
-        the ``linkage`` function documentation for more information
+        the `linkage` function documentation for more information
         on its structure.
 
     See Also
     --------
-    linkage
+    linkage: for advanced creation of hierarchical clusterings.
+    scipy.spatial.distance.pdist : pairwise distance metrics
 
     """
     return linkage(y, method='complete', metric='euclidean')
@@ -297,12 +301,12 @@ def average(y):
     -------
     Z : ndarray
         A linkage matrix containing the hierarchical clustering. See
-        the ``linkage`` function documentation for more information
-        on its structure.
+        `linkage` for more information on its structure.
 
     See Also
     --------
     linkage: for advanced creation of hierarchical clusterings.
+    scipy.spatial.distance.pdist : pairwise distance metrics
 
     """
     return linkage(y, method='average', metric='euclidean')
@@ -312,7 +316,7 @@ def weighted(y):
     """
     Performs weighted/WPGMA linkage on the condensed distance matrix.
 
-    See ``linkage`` for more information on the return
+    See `linkage` for more information on the return
     structure and algorithm.
 
     Parameters
@@ -325,12 +329,12 @@ def weighted(y):
     -------
     Z : ndarray
         A linkage matrix containing the hierarchical clustering. See
-        the ``linkage`` function documentation for more information
-        on its structure.
+        `linkage` for more information on its structure.
 
     See Also
     --------
     linkage : for advanced creation of hierarchical clusterings.
+    scipy.spatial.distance.pdist : pairwise distance metrics
 
     """
     return linkage(y, method='weighted', metric='euclidean')
@@ -340,27 +344,25 @@ def centroid(y):
     """
     Performs centroid/UPGMC linkage.
 
-    See ``linkage`` for more information on the return structure
-    and algorithm.
+    See `linkage` for more information on the input matrix,
+    return structure, and algorithm.
 
     The following are common calling conventions:
 
     1. ``Z = centroid(y)``
 
        Performs centroid/UPGMC linkage on the condensed distance
-       matrix ``y``.  See ``linkage`` for more information on the return
-       structure and algorithm.
+       matrix ``y``.
 
     2. ``Z = centroid(X)``
 
        Performs centroid/UPGMC linkage on the observation matrix ``X``
-       using Euclidean distance as the distance metric. See ``linkage``
-       for more information on the return structure and algorithm.
+       using Euclidean distance as the distance metric.
 
     Parameters
     ----------
     y : ndarray
-        A condensed or redundant distance matrix. A condensed
+        A condensed distance matrix. A condensed
         distance matrix is a flat array containing the upper
         triangular of the distance matrix. This is the form that
         ``pdist`` returns. Alternatively, a collection of
@@ -371,7 +373,7 @@ def centroid(y):
     -------
     Z : ndarray
         A linkage matrix containing the hierarchical clustering. See
-        the ``linkage`` function documentation for more information
+        the `linkage` function documentation for more information
         on its structure.
 
     See Also
@@ -386,7 +388,7 @@ def median(y):
     """
     Performs median/WPGMC linkage.
 
-    See ``linkage`` for more information on the return structure
+    See `linkage` for more information on the return structure
     and algorithm.
 
      The following are common calling conventions:
@@ -400,16 +402,16 @@ def median(y):
      2. ``Z = median(X)``
 
         Performs median/WPGMC linkage on the observation matrix ``X``
-        using Euclidean distance as the distance metric. See linkage
+        using Euclidean distance as the distance metric. See `linkage`
         for more information on the return structure and algorithm.
 
     Parameters
     ----------
     y : ndarray
-        A condensed or redundant distance matrix. A condensed
+        A condensed distance matrix. A condensed
         distance matrix is a flat array containing the upper
         triangular of the distance matrix. This is the form that
-        ``pdist`` returns. Alternatively, a collection of
+        ``pdist`` returns.  Alternatively, a collection of
         m observation vectors in n dimensions may be passed as
         a m by n array.
 
@@ -421,6 +423,7 @@ def median(y):
     See Also
     --------
     linkage: for advanced creation of hierarchical clusterings.
+    scipy.spatial.distance.pdist : pairwise distance metrics
 
     """
     return linkage(y, method='median', metric='euclidean')
@@ -428,41 +431,41 @@ def median(y):
 
 def ward(y):
     """
-    Performs Ward's linkage on a condensed or redundant distance matrix.
+    Performs Ward's linkage on a condensed distance matrix.
 
-    See linkage for more information on the return structure
+    See `linkage` for more information on the return structure
     and algorithm.
 
     The following are common calling conventions:
 
     1. ``Z = ward(y)``
-       Performs Ward's linkage on the condensed distance matrix ``Z``. See
-       linkage for more information on the return structure and
-       algorithm.
+       Performs Ward's linkage on the condensed distance matrix ``y``.
 
     2. ``Z = ward(X)``
        Performs Ward's linkage on the observation matrix ``X`` using
-       Euclidean distance as the distance metric. See linkage for more
-       information on the return structure and algorithm.
+       Euclidean distance as the distance metric.
 
     Parameters
     ----------
     y : ndarray
-        A condensed or redundant distance matrix. A condensed
+        A condensed distance matrix. A condensed
         distance matrix is a flat array containing the upper
         triangular of the distance matrix. This is the form that
-        ``pdist`` returns. Alternatively, a collection of
+        ``pdist`` returns.  Alternatively, a collection of
         m observation vectors in n dimensions may be passed as
         a m by n array.
 
     Returns
     -------
     Z : ndarray
-        The hierarchical clustering encoded as a linkage matrix.
+        The hierarchical clustering encoded as a linkage matrix. See
+        `linkage` for more information on the return structure and
+        algorithm.
 
     See Also
     --------
     linkage: for advanced creation of hierarchical clusterings.
+    scipy.spatial.distance.pdist : pairwise distance metrics
 
     """
     return linkage(y, method='ward', metric='euclidean')
@@ -470,15 +473,18 @@ def ward(y):
 
 def linkage(y, method='single', metric='euclidean'):
     """
-    Performs hierarchical/agglomerative clustering on the condensed
-    distance matrix y.
+    Performs hierarchical/agglomerative clustering.
 
-    y must be a :math:`{n \\choose 2}` sized
+    The input y may be either a 1d compressed distance matrix
+    or a 2d array of observation vectors.
+
+    If y is a 1d compressed distance matrix,
+    then y must be a :math:`{n \\choose 2}` sized
     vector where n is the number of original observations paired
     in the distance matrix. The behavior of this function is very
     similar to the MATLAB linkage function.
 
-    An :math:`(n-1)` by 4  matrix ``Z`` is returned. At the
+    A :math:`(n-1)` by 4 matrix ``Z`` is returned. At the
     :math:`i`-th iteration, clusters with indices ``Z[i, 0]`` and
     ``Z[i, 1]`` are combined to form cluster :math:`n + i`. A
     cluster with an index less than :math:`n` corresponds to one of
@@ -599,75 +605,84 @@ def linkage(y, method='single', metric='euclidean'):
     Parameters
     ----------
     y : ndarray
-        A condensed or redundant distance matrix. A condensed distance matrix
+        A condensed distance matrix. A condensed distance matrix
         is a flat array containing the upper triangular of the distance matrix.
         This is the form that ``pdist`` returns. Alternatively, a collection of
-        :math:`m` observation vectors in n dimensions may be passed as an
-        :math:`m` by :math:`n` array.
+        :math:`m` observation vectors in :math:`n` dimensions may be passed as an
+        :math:`m` by :math:`n` array. All elements of the condensed distance matrix
+        must be finite, i.e. no NaNs or infs.
     method : str, optional
         The linkage algorithm to use. See the ``Linkage Methods`` section below
         for full descriptions.
     metric : str or function, optional
         The distance metric to use in the case that y is a collection of
-        observation vectors; ignored otherwise. See the ``distance.pdist``
+        observation vectors; ignored otherwise. See the ``pdist``
         function for a list of valid distance metrics. A custom distance
-        function can also be used. See the ``distance.pdist`` function for
-        details.
+        function can also be used.
 
     Returns
     -------
     Z : ndarray
         The hierarchical clustering encoded as a linkage matrix.
 
+    Notes
+    -----
+    1. For method 'single' an optimized algorithm based on minimum spanning
+       tree is implemented. It has time complexity :math:`O(n^2)`.
+       For methods 'complete', 'average', 'weighted' and 'ward' an algorithm
+       called nearest-neighbors chain is implemented. It also has time
+       complexity :math:`O(n^2)`.
+       For other methods a naive algorithm is implemented with :math:`O(n^3)`
+       time complexity.
+       All algorithms use :math:`O(n^2)` memory.
+       Refer to [1]_ for details about the algorithms.
+    2. Methods 'centroid', 'median' and 'ward' are correctly defined only if
+       Euclidean pairwise metric is used. If `y` is passed as precomputed
+       pairwise distances, then it is a user responsibility to assure that
+       these distances are in fact Euclidean, otherwise the produced result
+       will be incorrect.
+
+    See Also
+    --------
+    scipy.spatial.distance.pdist : pairwise distance metrics
+
+    References
+    ----------
+    .. [1] Daniel Mullner, "Modern hierarchical, agglomerative clustering
+           algorithms", :arXiv:`1109.2378v1`.
     """
-    if not isinstance(method, string_types):
-        raise TypeError("Argument 'method' must be a string.")
+    if method not in _LINKAGE_METHODS:
+        raise ValueError("Invalid method: {0}".format(method))
 
     y = _convert_to_double(np.asarray(y, order='c'))
 
-    s = y.shape
-    if len(s) == 1:
+    if y.ndim == 1:
         distance.is_valid_y(y, throw=True, name='y')
-        d = distance.num_obs_y(y)
-        if method not in _cpy_non_euclid_methods:
-            raise ValueError("Valid methods when the raw observations are "
-                             "omitted are 'single', 'complete', 'weighted', "
-                             "and 'average'.")
-        # Since the C code does not support striding using strides.
         [y] = _copy_arrays_if_base_present([y])
+    elif y.ndim == 2:
+        if method in _EUCLIDEAN_METHODS and metric != 'euclidean':
+            raise ValueError("Method '{0}' requires the distance metric "
+                             "to be Euclidean".format(method))
+        if y.shape[0] == y.shape[1] and np.allclose(np.diag(y), 0):
+            if np.all(y >= 0) and np.allclose(y, y.T):
+                _warning('The symmetric non-negative hollow observation '
+                         'matrix looks suspiciously like an uncondensed '
+                         'distance matrix')
+        y = distance.pdist(y, metric)
+    else:
+        raise ValueError("`y` must be 1 or 2 dimensional.")
 
-        Z = np.zeros((d - 1, 4))
+    if not np.all(np.isfinite(y)):
+        raise ValueError("The condensed distance matrix must contain only finite values.")
 
-        if method == 'single':
-            _hierarchy.slink(y, Z, int(d))
-        else:
-            _hierarchy.linkage(y, Z, int(d),
-                               int(_cpy_non_euclid_methods[method]))
-
-    elif len(s) == 2:
-        X = y
-        n = s[0]
-        if method not in _cpy_linkage_methods:
-            raise ValueError('Invalid method: %s' % method)
-        if method in _cpy_non_euclid_methods:
-            dm = distance.pdist(X, metric)
-            Z = np.zeros((n - 1, 4))
-
-            if method == 'single':
-                _hierarchy.slink(dm, Z, n)
-            else:
-                _hierarchy.linkage(dm, Z, n,
-                                   int(_cpy_non_euclid_methods[method]))
-
-        elif method in _cpy_euclid_methods:
-            if metric != 'euclidean':
-                raise ValueError(("Method '%s' requires the distance metric "
-                                 "to be euclidean") % method)
-            dm = distance.pdist(X, metric)
-            Z = np.zeros((n - 1, 4))
-            _hierarchy.linkage(dm, Z, n,
-                               int(_cpy_euclid_methods[method]))
-    return Z
+    n = int(distance.num_obs_y(y))
+    method_code = _LINKAGE_METHODS[method]
+    if method == 'single':
+        return _hierarchy.mst_single_linkage(y, n)
+    elif method in ['complete', 'average', 'weighted', 'ward']:
+        return _hierarchy.nn_chain(y, n, method_code)
+    else:
+        return _hierarchy.fast_linkage(y, n, method_code)
 
 
 class ClusterNode:
@@ -677,9 +692,24 @@ class ClusterNode:
     Leaf nodes correspond to original observations, while non-leaf nodes
     correspond to non-singleton clusters.
 
-    The to_tree function converts a matrix returned by the linkage
+    The `to_tree` function converts a matrix returned by the linkage
     function into an easy-to-use tree representation.
 
+    All parameter names are also attributes.
+
+    Parameters
+    ----------
+    id : int
+        The node id.
+    left : ClusterNode instance, optional
+        The left child tree node.
+    right : ClusterNode instance, optional
+        The right child tree node.
+    dist : float, optional
+        Distance for this cluster in the linkage matrix.
+    count : int, optional
+        The number of samples in this cluster.
+
     See Also
     --------
     to_tree : for converting a linkage matrix ``Z`` into a tree object.
@@ -812,9 +842,10 @@ class ClusterNode:
         ----------
         func : function
             Applied to each leaf ClusterNode object in the pre-order traversal.
-            Given the i'th leaf node in the pre-ordeR traversal ``n[i]``, the
-            result of func(n[i]) is stored in L[i]. If not provided, the index
-            of the original observation to which the node corresponds is used.
+            Given the ``i``-th leaf node in the pre-order traversal ``n[i]``, the
+            result of ``func(n[i])`` is stored in ``L[i]``. If not provided,
+            the index of the original observation to which the node
+            corresponds is used.
 
         Returns
         -------
@@ -822,7 +853,6 @@ class ClusterNode:
             The pre-order traversal.
 
         """
-
         # Do a preorder traversal, caching the result. To avoid having to do
         # recursion, we'll store the previous index we've visited in a vector.
         n = self.count
@@ -971,39 +1001,60 @@ def cut_tree(Z, n_clusters=None, height=None):
 
 def to_tree(Z, rd=False):
     """
-    Converts a hierarchical clustering encoded in the matrix ``Z`` (by
-    linkage) into an easy-to-use tree object.
+    Converts a linkage matrix into an easy-to-use tree object.
 
-    The reference r to the root ClusterNode object is returned.
+    The reference to the root `ClusterNode` object is returned (by default).
 
-    Each ClusterNode object has a left, right, dist, id, and count
-    attribute. The left and right attributes point to ClusterNode objects
-    that were combined to generate the cluster. If both are None then
-    the ClusterNode object is a leaf node, its count must be 1, and its
-    distance is meaningless but set to 0.
+    Each `ClusterNode` object has a ``left``, ``right``, ``dist``, ``id``,
+    and ``count`` attribute. The left and right attributes point to
+    ClusterNode objects that were combined to generate the cluster.
+    If both are None then the `ClusterNode` object is a leaf node, its count
+    must be 1, and its distance is meaningless but set to 0.
 
-    Note: This function is provided for the convenience of the library
+    *Note: This function is provided for the convenience of the library
     user. ClusterNodes are not used as input to any of the functions in this
-    library.
+    library.*
 
     Parameters
     ----------
     Z : ndarray
-        The linkage matrix in proper form (see the ``linkage``
+        The linkage matrix in proper form (see the `linkage`
         function documentation).
     rd : bool, optional
-        When False, a reference to the root ClusterNode object is
-        returned.  Otherwise, a tuple (r,d) is returned. ``r`` is a
-        reference to the root node while ``d`` is a dictionary
-        mapping cluster ids to ClusterNode references. If a cluster id is
-        less than n, then it corresponds to a singleton cluster
-        (leaf node). See ``linkage`` for more information on the
-        assignment of cluster ids to clusters.
+        When False (default), a reference to the root `ClusterNode` object is
+        returned.  Otherwise, a tuple ``(r, d)`` is returned. ``r`` is a
+        reference to the root node while ``d`` is a list of `ClusterNode`
+        objects - one per original entry in the linkage matrix plus entries
+        for all clustering steps.  If a cluster id is
+        less than the number of samples ``n`` in the data that the linkage
+        matrix describes, then it corresponds to a singleton cluster (leaf
+        node).
+        See `linkage` for more information on the assignment of cluster ids
+        to clusters.
 
     Returns
     -------
-    L : list
-        The pre-order traversal.
+    tree : ClusterNode or tuple (ClusterNode, list of ClusterNode)
+        If ``rd`` is False, a `ClusterNode`.
+        If ``rd`` is True, a list of length ``2*n - 1``, with ``n`` the number
+        of samples.  See the description of `rd` above for more details.
+
+    See Also
+    --------
+    linkage, is_valid_linkage, ClusterNode
+
+    Examples
+    --------
+    >>> from scipy.cluster import hierarchy
+    >>> x = np.random.rand(10).reshape(5, 2)
+    >>> Z = hierarchy.linkage(x)
+    >>> hierarchy.to_tree(Z)
+    <scipy.cluster.hierarchy.ClusterNode object at ...
+    >>> rootnode, nodelist = hierarchy.to_tree(Z, rd=True)
+    >>> rootnode
+    <scipy.cluster.hierarchy.ClusterNode object at ...
+    >>> len(nodelist)
+    9
 
     """
     Z = np.asarray(Z, order='c')
@@ -1033,7 +1084,7 @@ def to_tree(Z, rd=False):
                               'is used before it is formed. See row %d, '
                               'column 1') % fj)
         nd = ClusterNode(i + n, d[fi], d[fj], Z[i, 2])
-        #          ^ id   ^ left ^ right ^ dist
+        #                 ^ id   ^ left ^ right ^ dist
         if Z[i, 3] != nd.count:
             raise ValueError(('Corrupt matrix Z. The count Z[%d,3] is '
                               'incorrect.') % i)
@@ -1088,7 +1139,7 @@ def cophenet(Z, Y=None):
     Returns
     -------
     c : ndarray
-        The cophentic correlation distance (if ``y`` is passed).
+        The cophentic correlation distance (if ``Y`` is passed).
     d : ndarray
         The cophenetic distance matrix in condensed form. The
         :math:`ij` th entry is the cophenetic distance between
@@ -1125,10 +1176,7 @@ def cophenet(Z, Y=None):
 
 def inconsistent(Z, d=2):
     r"""
-    Calculates inconsistency statistics on a linkage.
-
-    Note: This function behaves similarly to the MATLAB(TM)
-    inconsistent function.
+    Calculates inconsistency statistics on a linkage matrix.
 
     Parameters
     ----------
@@ -1152,6 +1200,11 @@ def inconsistent(Z, d=2):
 
         .. math:: \frac{\mathtt{Z[i,2]} - \mathtt{R[i,0]}} {R[i,1]}
 
+    Notes
+    -----
+    This function behaves similarly to the MATLAB(TM) ``inconsistent``
+    function.
+
     """
     Z = np.asarray(Z, order='c')
 
@@ -1182,9 +1235,9 @@ def from_mlab_linkage(Z):
      * the indices are converted from ``1..N`` to ``0..(N-1)`` form,
        and
 
-     * a fourth column Z[:,3] is added where Z[i,3] is represents the
+     * a fourth column ``Z[:,3]`` is added where ``Z[i,3]`` represents the
        number of original observations (leaves) in the non-singleton
-       cluster i.
+       cluster ``i``.
 
     This function is useful when loading in linkages from legacy data
     files generated by MATLAB.
@@ -1197,7 +1250,7 @@ def from_mlab_linkage(Z):
     Returns
     -------
     ZS : ndarray
-        A linkage matrix compatible with this library.
+        A linkage matrix compatible with ``scipy.cluster.hierarchy``.
 
     """
     Z = np.asarray(Z, dtype=np.double, order='c')
@@ -1236,7 +1289,7 @@ def to_mlab_linkage(Z):
     Parameters
     ----------
     Z : ndarray
-        A linkage matrix generated by this library.
+        A linkage matrix generated by ``scipy.cluster.hierarchy``.
 
     Returns
     -------
@@ -1289,7 +1342,7 @@ def is_monotonic(Z):
 def is_valid_im(R, warning=False, throw=False, name=None):
     """Returns True if the inconsistency matrix passed is valid.
 
-    It must be a :math:`n` by 4 numpy array of doubles. The standard
+    It must be a :math:`n` by 4 array of doubles. The standard
     deviations ``R[:,1]`` must be nonnegative. The link counts
     ``R[:,2]`` must be positive and no greater than :math:`n-1`.
 
@@ -1379,8 +1432,8 @@ def is_valid_linkage(Z, warning=False, throw=False, name=None):
         When True, throws a Python exception if the linkage
         matrix passed is invalid.
     name : str, optional
-           This string refers to the variable name of the invalid
-           linkage matrix.
+        This string refers to the variable name of the invalid
+        linkage matrix.
 
     Returns
     -------
@@ -1517,7 +1570,7 @@ def correspond(Z, Y):
 def fcluster(Z, t, criterion='inconsistent', depth=2, R=None, monocrit=None):
     """
     Forms flat clusters from the hierarchical clustering defined by
-    the linkage matrix ``Z``.
+    the given linkage matrix.
 
     Parameters
     ----------
@@ -1584,8 +1637,8 @@ def fcluster(Z, t, criterion='inconsistent', depth=2, R=None, monocrit=None):
     Returns
     -------
     fcluster : ndarray
-        An array of length n. T[i] is the flat cluster number to
-        which original observation i belongs.
+        An array of length ``n``. ``T[i]`` is the flat cluster number to
+        which original observation ``i`` belongs.
 
     """
     Z = np.asarray(Z, order='c')
@@ -1636,8 +1689,9 @@ def fclusterdata(X, t, criterion='inconsistent',
     and forms flat clusters using the inconsistency method with `t` as the
     cut-off threshold.
 
-    A one-dimensional array T of length n is returned. T[i] is the index
-    of the flat cluster to which the original observation i belongs.
+    A one-dimensional array ``T`` of length ``n`` is returned. ``T[i]`` is
+    the index of the flat cluster to which the original observation ``i``
+    belongs.
 
     Parameters
     ----------
@@ -1651,7 +1705,7 @@ def fclusterdata(X, t, criterion='inconsistent',
         cluster formation algorithms. See `fcluster` for descriptions.
     metric : str, optional
         The distance metric for calculating pairwise distances. See
-        `distance.pdist` for descriptions and linkage to verify
+        ``distance.pdist`` for descriptions and linkage to verify
         compatibility with the linkage method.
     depth : int, optional
         The maximum depth for the inconsistency calculation. See
@@ -1670,9 +1724,13 @@ def fclusterdata(X, t, criterion='inconsistent',
         A vector of length n. T[i] is the flat cluster number to
         which original observation i belongs.
 
+    See Also
+    --------
+    scipy.spatial.distance.pdist : pairwise distance metrics
+
     Notes
     -----
-    This function is similar to the MATLAB function clusterdata.
+    This function is similar to the MATLAB function ``clusterdata``.
 
     """
     X = np.asarray(X, order='c', dtype=np.double)
@@ -1702,7 +1760,7 @@ def leaves_list(Z):
     ----------
     Z : ndarray
         The hierarchical clustering encoded as a matrix.  `Z` is
-        a linkage matrix.  See ``linkage`` for more information.
+        a linkage matrix.  See `linkage` for more information.
 
     Returns
     -------
@@ -1980,11 +2038,12 @@ def dendrogram(Z, p=30, truncate_mode=None, color_threshold=None,
 
     The dendrogram illustrates how each cluster is
     composed by drawing a U-shaped link between a non-singleton
-    cluster and its children. The height of the top of the U-link is
-    the distance between its children clusters. It is also the
+    cluster and its children.  The top of the U-link indicates a
+    cluster merge.  The two legs of the U-link indicate which clusters
+    were merged.  The length of the two legs of the U-link represents
+    the distance between the child clusters.  It is also the
     cophenetic distance between original observations in the two
-    children clusters. It is expected that the distances in Z[:,2] be
-    monotonic, otherwise crossings appear in the dendrogram.
+    children clusters.
 
     Parameters
     ----------
@@ -2000,21 +2059,23 @@ def dendrogram(Z, p=30, truncate_mode=None, color_threshold=None,
         large. Truncation is used to condense the dendrogram. There
         are several modes:
 
-        ``None/'none'``
-          No truncation is performed (Default).
+        ``None``
+          No truncation is performed (default).
+          Note: ``'none'`` is an alias for ``None`` that's kept for
+          backward compatibility.
 
         ``'lastp'``
-          The last ``p`` non-singleton formed in the linkage are the only
-          non-leaf nodes in the linkage; they correspond to rows
+          The last ``p`` non-singleton clusters formed in the linkage are the
+          only non-leaf nodes in the linkage; they correspond to rows
           ``Z[n-p-2:end]`` in ``Z``. All other non-singleton clusters are
           contracted into leaf nodes.
 
-        ``'mlab'``
-          This corresponds to MATLAB(TM) behavior. (not implemented yet)
-
-        ``'level'/'mtica'``
+        ``'level'``
           No more than ``p`` levels of the dendrogram tree are displayed.
-          This corresponds to Mathematica(TM) behavior.
+          A "level" includes all nodes with ``p`` merges from the last merge.
+
+          Note: ``'mtica'`` is an alias for ``'level'`` that's kept for
+          backward compatibility.
 
     color_threshold : double, optional
         For brevity, let :math:`t` be the ``color_threshold``.
@@ -2186,6 +2247,11 @@ def dendrogram(Z, p=30, truncate_mode=None, color_threshold=None,
     --------
     linkage, set_link_color_palette
 
+    Notes
+    -----
+    It is expected that the distances in ``Z[:,2]`` be monotonic, otherwise
+    crossings appear in the dendrogram.
+
     Examples
     --------
     >>> from scipy.cluster import hierarchy
@@ -2236,13 +2302,18 @@ def dendrogram(Z, p=30, truncate_mode=None, color_threshold=None,
         raise TypeError('The second argument must be a number')
 
     if truncate_mode not in ('lastp', 'mlab', 'mtica', 'level', 'none', None):
+        # 'mlab' and 'mtica' are kept working for backwards compat.
         raise ValueError('Invalid truncation mode.')
 
     if truncate_mode == 'lastp' or truncate_mode == 'mlab':
         if p > n or p == 0:
             p = n
 
-    if truncate_mode == 'mtica' or truncate_mode == 'level':
+    if truncate_mode == 'mtica':
+        # 'mtica' is an alias
+        truncate_mode = 'level'
+
+    if truncate_mode == 'level':
         if p <= 0:
             p = np.inf
 
@@ -2419,10 +2490,10 @@ def _dendrogram_calculate_info(Z, p, truncate_mode,
         raise ValueError("Invalid root cluster index i.")
 
     if truncate_mode == 'lastp':
-        # If the node is a leaf node but corresponds to a non-single cluster,
+        # If the node is a leaf node but corresponds to a non-singleton cluster,
         # its label is either the empty string or the number of original
         # observations belonging to cluster i.
-        if 2 * n - p > i >= n:
+        if 2*n - p > i >= n:
             d = Z[i - n, 2]
             _append_nonsingleton_leaf_node(Z, p, n, level, lvs, ivl,
                                            leaf_label_func, i, labels,
@@ -2434,7 +2505,7 @@ def _dendrogram_calculate_info(Z, p, truncate_mode,
             _append_singleton_leaf_node(Z, p, n, level, lvs, ivl,
                                         leaf_label_func, i, labels)
             return (iv + 5.0, 10.0, 0.0, 0.0)
-    elif truncate_mode in ('mtica', 'level'):
+    elif truncate_mode == 'level':
         if i > n and level > p:
             d = Z[i - n, 2]
             _append_nonsingleton_leaf_node(Z, p, n, level, lvs, ivl,
@@ -2448,13 +2519,11 @@ def _dendrogram_calculate_info(Z, p, truncate_mode,
                                         leaf_label_func, i, labels)
             return (iv + 5.0, 10.0, 0.0, 0.0)
     elif truncate_mode in ('mlab',):
-        pass
+        msg = "Mode 'mlab' is deprecated in scipy 0.19.0 (it never worked)."
+        warnings.warn(msg, DeprecationWarning)
 
     # Otherwise, only truncate if we have a leaf node.
     #
-    # If the truncate_mode is mlab, the linkage has been modified
-    # with the truncated tree.
-    #
     # Only place leaves if they correspond to original observations.
     if i < n:
         _append_singleton_leaf_node(Z, p, n, level, lvs, ivl,
@@ -2633,13 +2702,19 @@ def is_isomorphic(T1, T2):
     if T1S[0] != T2S[0]:
         raise ValueError('T1 and T2 must have the same number of elements.')
     n = T1S[0]
-    d = {}
+    d1 = {}
+    d2 = {}
     for i in xrange(0, n):
-        if T1[i] in d:
-            if d[T1[i]] != T2[i]:
+        if T1[i] in d1:
+            if not T2[i] in d2:
                 return False
+            if d1[T1[i]] != T2[i] or d2[T2[i]] != T1[i]:
+                return False
+        elif T2[i] in d2:
+            return False
         else:
-            d[T1[i]] = T2[i]
+            d1[T1[i]] = T2[i]
+            d2[T2[i]] = T1[i]
     return True
 
 
@@ -2682,7 +2757,7 @@ def maxinconsts(Z, R):
     ----------
     Z : ndarray
         The hierarchical clustering encoded as a matrix. See
-        ``linkage`` for more information.
+        `linkage` for more information.
     R : ndarray
         The inconsistency matrix.
 
@@ -2780,7 +2855,7 @@ def leaders(Z, T):
     ----------
     Z : ndarray
         The hierarchical clustering encoded as a matrix. See
-        ``linkage`` for more information.
+        `linkage` for more information.
     T : ndarray
         The flat cluster assignment vector.
 
